-- -*-coding: utf-8;-*-

--[[
-- ************************************************************************
-- BuddyName.lua
-- ************************************************************************
-- This addon is released under GPL v3. All additionally used libraries are
-- released under the license that accompanies them.
--]]

BuddyName = LibStub("AceAddon-3.0"):NewAddon("BuddyName",
                                             "AceConsole-3.0",
                                             "AceHook-3.0",
                                             "AceTimer-3.0",
                                             "AceComm-3.0");

local L = LibStub("AceLocale-3.0"):GetLocale("BuddyName");

local Compressor = LibStub:GetLibrary("LibCompress");

local chatChannels = {
	 ["WHISPER"]      = true,
	 ["PARTY"]        = true,
	 ["GUILD"]        = true,
	 ["YELL"]         = true,
	 ["PARTY"]        = true,
	 ["RAID"]         = true,
	 ["BATTLEGROUND"] = true,
	 ["CHANNEL"]      = false,
}

local slashCommands = {
	 ["bn"]        = true,
	 ["buddyname"] = true,
	 ["BuddyName"] = true,
}

-- perhaps used in future versions
local registerEvents = {
	 ["GUILD_JOIN"] = false,
}

-- Constants used for synchronization
local COMM = {
	 SYNC_PREFIX = "BuddyName",
	 PROTOCOL = 2,
	 SEP = { -- separator
			SECTION = "|",
			ENTRY   = ";",
			FIELD   = ":",
	 },
	 MSG_TYPE = {
			ENTRY_SET        = 1,
			HASH_PROPAGATION = 2,
			DATA_REQUEST     = 3,
			UNKNOWN          = -1,
	 },
	 HEADER = {
			PREFIX="BN",
			COMPRESSION = {
				 NONE        = 0,
				 LIBCOMPRESS = 1,
			},
	 },
	 PATTERNS = {}, -- filled in below
};
COMM.PATTERNS = {
	 SEARCH = {
			HEADER   = "^"..COMM.HEADER.PREFIX..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.SECTION,
			MSG_TYPE = "^(%d+)"..COMM.SEP.ENTRY,
	 },
	 BUILD = {
			HEADER = COMM.HEADER.PREFIX..COMM.SEP.ENTRY..COMM.PROTOCOL..COMM.SEP.ENTRY..COMM.HEADER.COMPRESSION.LIBCOMPRESS..COMM.SEP.SECTION,
			ENTRY_SET_STUB = COMM.MSG_TYPE.ENTRY_SET,
	 },
};

--[[ Local variables ]]
local BN_DB_Version = 2; -- Database format
local BN_CommVersion = 1; -- Message format communication protocol
local BN_Active = false; -- Is this addon active? Modified by OnEnable and OnDisable
local BN_EventFrame = nil; -- Used to register events and process them. Created in OnInitialize
local currentRealm = nil;
local BN_Timer = {}; -- table for storing active timers
local warnedAboutUpdate = false; -- whether the user has been warned about an existing update this session
local compressType = 1; -- 0 means no compression, 1 means usage of LibCompression
local commActive = false; -- debug, remove me later

--[[ -- -------------------------------- -- ]]
--[[ -- Basic functions, no dependenices -- ]]
--[[ -- -------------------------------- -- ]]

--[[ Creates a new numeric timestamp for the current time and date
--   Returns:
--     1. Numeric: timestamp in format YYYYMMDDHHmm
--]]
function BuddyName:GetTimestamp()
	 local _,month,day,year = CalendarGetDate();
	 local hour,minute = GetGameTime();

	 local timestamp = minute + 100 * (hour + 100 * (day + 100 * (month + 100 * year)))

	 -- self:Print("Debug Timestamp: "..timestamp);

	 return timestamp
end

--[[ Concatenates an arbitrary number of arguments to one string.
--   Only used for debugging purposes.
--   Returns:
--     1. Concatenated version of all supplied arguments
--]]
local function allToString(...)
	 return strjoin(" ", tostringall(...));
end

local function PlayerIsInGuild()
	 local guild = GetGuildInfo("player");
	 if (guild) then
			return true;
	 else
			return false;
	 end
end

--[[ -- -------------------------- -- ]]
--[[ -- Database-related functions -- ]]
--[[ -- -------------------------- -- ]]

--[[ Transforms old DB entries to newer versions
--   Returns: nil
--]]
function BuddyName:UpdateDB()
	 if (self.db.realm.version==1) then
			self:Print(string.format(L["UPDATING_DB"], 2));

			-- perform updates to v2
			for p,v in pairs(self.db.realm.char) do
				 if (self.db.realm.char[p]) then
						self.db.realm.char[p] = {
							 buddy = v,
							 guild = "", -- TODO
							 timestamp = BuddyName:GetTimestamp(),
						}
				 end
			end

			self.db.realm.version=2
			self:Print(L["UPDATE_OK"])
	 end

	 --[[
	 if (self.db.realm.version==2) then
			self:Print(string.format(L["UPDATING_DB"], 3));
			-- perform update to v3 here
			self.db.realm.version=3
			self:Print(L["UPDATE_OK"])
	 end
	 ]]
end

--[[ Checks whether the given charname has a buddy name in our DB
--   Returns:
--     1. Boolean: true if author was found in DB, false else
--     2. String: Buddy name of charname, or nil if it doesn't exist or equals empty string
--]]
function BuddyName:IsInDB(charname)
	 if (charname~=nil and charname~="") then
			if (self.db.realm.char and -- simplify? test!
					self.db.realm.char[charname] and
					self.db.realm.char[charname].buddy and
					self.db.realm.char[charname].buddy~="")
			then
				 return true,self.db.realm.char[charname].buddy;
			end
	 end

	 return false, nil
end

--[[ Adds a new entry to the DB or updates an existing one.
--   If a timestamp is provided, then this will be used.
--   Returns:
--     1. Boolean: true if successul, false else
--]]
function BuddyName:addToDB(charname, buddyname, timestamp)
	 local ts;
	 if (timestamp) then
			ts = timestamp;
	 else
			ts = BuddyName:GetTimestamp();
	 end
	 self.db.realm.char[charname]= {
			["buddy"] = buddyname,
			["guild"] = "", -- TODO
			["timestamp"] = ts
	 }
	 return true
end

-- returns true if updated, false if not updated
function BuddyName:UpdateIfNewer(charname, buddyname, timestamp)
	 if (charname~=nil and charname~="" and buddyname~=nil and timestamp~=nil) then
			if (not self.db.realm.char[charname]) then -- entry does not exist at all
				 self:Print(string.format("Added new char from synchronization: %s => %s", charname, buddyname)); -- debug
				 BuddyName:addToDB(charname,buddyname,timestamp);
				 return true;
			elseif (self.db.realm.char[charname].timestamp < timestamp) then -- Check if received timestamp is newer
				 self:Print(string.format("Updated char from synchronization: %s => %s", charname, buddyname)); -- debug
				 BuddyName:addToDB(charname,buddyname,timestamp);
				 return true;
			end
	 end
	 return false;
end

--[[ Deletes an entry from the DB if it exists. Depending on the deleteCompletely flag,
--   an existing entry may either be deleted completely (true) or just set to "" (false).
--   The latter leaves the actual entry intact, resulting in its spreading over the guild
--   synchronization mechanism. Therefore, it is possible to delete this entry from other
--   guild members DBs as well.
--   Returns:
--     1. Boolean: true if deletion was successful, false else
--     2. String: buddy name if deletion was successful, nil else (e.g. char was not in DB)
--]]
function BuddyName:deleteFromDB(charname, deleteCompletely)
	 -- is this char in the DB?
	 if (not BuddyName.db.realm.char[charname]) then
			return false,nil
	 end

	 -- perform deletion and set return value
	 local buddyname = BuddyName.db.realm.char[charname].buddy
	 if (deleteCompletely==true) then
			BuddyName.db.realm.char[charname]=nil
	 else
			BuddyName.db.realm.char[charname].buddy=""
			BuddyName.db.realm.char[charname].timestamp=BuddyName:GetTimestamp();
	 end

	 return true,buddyname
end

function BuddyName:FixDB(version)
	 if (version==2) then
			-- remove all buddy names with "|K...", as they result from broken RealID stuff
			for p,v in pairs(self.db.realm.char) do
				 local pstart = string.find(self.db.realm.char[p].buddy, "^|K");
				 if (pstart) then
						self.db.realm.char[p] = nil;
						self:Print(string.format("Removed broken entry for %s",p));
				 end
			end
	 end
end

--[[ -- -------------- -- ]]
--[[ -- Message output -- ]]
--[[ -- -------------- -- ]]

--[[ Hooked-in version of the AddMessage function.
--   Will check if message contains matching character link,
--   and insert buddy name if match was found.
--   Returns: nil
--]]
function BuddyName:AddMessage(frame, msg, ...)
	 local pstart,pend,sender,type = string.find(msg, "|Hplayer:(%a+):%d+:(%u+)[^|]*|h%[[^%]]+%]"); -- TODO: try to match against |h..|h instead of [..]
	 if (pstart) then -- something matching was found, else nil
			local replace,buddy = BuddyName:IsInDB(sender) -- char in DB?
			if (replace) then
				 if (chatChannels[type]==true or -- msg occured in one of the monitored channels
						string.find(msg,L["ANNOUNCE_ONLINE"]) or -- player has come online
						string.find(msg,L["ANNOUNCE_ACHIEVEMENT"]) ) -- player earned an achievement
				 then
						msg=string.gsub(msg, "(%[[^%]]*)("..sender..")([^%[]*%])", '%1%2:'..buddy..'%3',1); -- only replace first occurrence
				 end
			end
	 else
			pstart,pend,sender=string.find(msg, L["ANNOUNCE_OFFLINE"]);
			if (pstart) then
				 local replace,buddy = BuddyName:IsInDB(sender) -- char in DB?
				 if (replace) then
						msg=string.gsub(msg, L["ANNOUNCE_OFFLINE"], "%1"..buddy, 1);
				 end
			end
	 end
	 
	 -- call original AddMessage() function
	 self.hooks[frame]["AddMessage"](frame, msg, ...);
end

--[[ Event Handler that is currently not in use. Perhaps in future version agains.
--   Returns: nil
--]]
local function eventHandler(self, event, ...)
	 if (BN_Active == false) then
			return;
	 end

	 if (registerEvents[event]==true) then
			--print("Registered event of type "..event); -- debug
	 end
end

--[[ -- ----------------- -- ]]
--[[ -- AddOn interaction -- ]]
--[[ -- ----------------- -- ]]

--[[ Process slash commands (unexpected, eh?)
--   Returns: nil
--]]
function BuddyName:SlashProcessor(input)
	 -- 'input' contains whatever follows the slash command

	 -- split input string, ignore additional whitespace
	 local tbl = {}
	 for v in string.gmatch(input, "[^ ]+") do
			tinsert(tbl, v)
	 end

	 -- if no additional parameters => print out help text
	 if (#tbl == 0 or (tbl[1]~=nil and tbl[1]=="help")) then
			self:Print(L["HELP"])
			return;
	 end

	 local command = tremove(tbl,1);

	 -- Activate additional functionality
	 if (command=="on") then
			self:Enable()
			self:Print(L["ACTIVATED"]);
			return
	 end

	 -- Disable additional functionality
	 if (command=="off") then
			self:Disable()
			self:Print(L["DEACTIVATED"]);
			return
	 end

	 if (not BN_Active) then
			self:Print(L["NOT_ACTIVE"])
			return
	 end

	 -- add another person to the list
	 if (command=="add") then
			if (#tbl~=2) then
				 self:Print("Add: "..L["WRONG_NUMBER_OF_ARGUMENTS"])
			end
			BuddyName:addToDB(tbl[1], tbl[2]);
			self:Print(string.format(L["BUDDY_ADD_OK"], tbl[1], tbl[2]));
			return
	 end

	 -- add another person to the list
	 if (command=="del" or command=="rm") then
			if (#tbl~=1) then
				 self:Print("Del: "..L["WRONG_NUMBER_OF_ARGUMENTS"])
			end
			local successful,buddyname = BuddyName:deleteFromDB(tbl[1], false)
			if (not successful) then
				 self:Print(string.format(L["BUDDY_DEL_ERROR"], tbl[1]))
				 return
			end
			self:Print(string.format(L["BUDDY_DEL_OK"], tbl[1], buddyname));
			return
	 end

	 -- list chars from database
	 if (command=="list") then
			self:Print(string.format(L["BUDDY_LIST_HEAD"],currentRealm));

			for p,v in pairs(BuddyName.db.realm.char) do
				 local char = p;
				 local buddy = BuddyName.db.realm.char[p].buddy -- DB v2
				 if (buddy and buddy~="") then
						self:Print(string.format(L["BUDDY_LIST_ENTRY"], char, buddy));
				 end
			end

			-- TODO:
			-- Distinguish between all, guild, non-guild
			-- => Additional command arguments
			return
	 end

	 -- to test the communication properties
	 if (command=="comm") then
			if (commActive==true) then
				 commActive=false;
				 self:Print("Switched off communication");
			else
				 commActive=true;
				 self:Print("Switched on communication");
			end
			return;
	 end

	 -- to test the communication properties
	 if (command=="sync") then
			self:Print("Initiating guild broadcast");
	 end

	 self:Print(string.format(L["UNKNOWN_COMMAND"], command));

-- debug
--	 for v = 1,(#tbl) do
--			self:Print(ChatFrame1, tbl[v]);
--	 end
end

--[[ -- ------------------------- -- ]]
--[[ -- AddOn chat communication  -- ]]
--[[ -- ------------------------- -- ]]

-- extracts header from message string, leaves string intact.
-- returns nil if no matching header was found, else table consisting of header entries and header length
local function CommGetMsgHeader(msg)
	 -- TODO: only do this once in the header in the future
	 -- local headerPattern = "^"..COMM.HEADER.PREFIX..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.SECTION;
	 -- print("Debug header pattern: "..headerPattern); -- debug, remove me
	 -- local hstart,hlength,hprotocol,hcompression = string.find(headerPattern);

	 print("Debug header pattern: "..COMM.PATTERNS.SEARCH.HEADER); -- debug, remove me

	 local hstart,hlength,hprotocol,hcompression = string.find(COMM.PATTERNS.SEARCH.HEADER);
	 local header = nil;
	 if (hstart) then
			header = {
				 protocol = hprotocol,
				 compression = hcompression,
				 length = hlength,
			};
	 end

	 print("Header: "..allToString(header)); -- debug, remove me

	 return header;
end

-- determines type of message and splits up textual representation
-- according to message format definition.
-- returns table with type and data fields.
local function CommDeserializeMsg(msg)

	 local pstart,plength,msgType = string.find(msg, COMM.PATTERNS.SEARCH.MSG_TYPE);

	 if (not pstart) then
			return { type = COMM.MSG_TYPE.UNKNOWN };
	 end

	 local splitMsg = {
			type  = msgType,
			entry = nil,
	 };
	 local entries={ strsplit(COMM.SEP.ENTRY, strsub(msg,plength+1)) };

	 if (msgType==COMM.MSG_TYPE.ENTRY_SET) then
			splitMsg.entry = {};
			for i=1, #entries do
				 if (entries[i]~=nil and entries[i]~="") then
						local entry = { strsplit(COMM.SEP.FIELD, entries[i]); };
						if (#entry ~= 3) then
							 print("Malformed ENTRY_SET message");
						else
							 entry[3] = tonumber(entry[3]);
							 if (entry[3]==nil) then
									print("Malformed ENTRY_SET message");
							 else
									tinsert(splitMsg.entry, entry);
							 end
						end
				 end
			end
	 elseif (msgType==COMM.MSG_TYPE.HASH_PROPAGATION) then
			if (#entries ~= 4) then
				 print("Malformed HASH_PROPAGATION message");
			else
				 splitMsg.entry = {
						dataDomain = tonumber(entries[1]),
						hash       =          entries[2],
						entries    = tonumber(entries[3]),
						timestamp  = tonumber(entries[4]),
				 };
			end
	 elseif (msgType==COMM.MSG_TYPE.DATA_REQUEST) then
			if (#entries ~= 3) then
				 print("Malformed DATA_REQUEST message");
			else
				 splitMsg.entry = {
						dataDomain = tonumber(entries[1]),
						partial    = tonumber(entries[2]),
						response   = tonumber(entries[3]),
				 };
			end
	 else
			return { type = COMM.MSG_TYPE.UNKNOWN };
	 end

	 return splitMsg;
end

-- Splits the message up into its various parts
-- Returns a table consisting of the header, and one entry per following message.
-- If malformed message then returned table equals nil.
local function CommSplitMsg(rawMsg)
	 local header = CommGetMsgHeader(rawMsg);
	 if (header == nil) then -- no valid header was found
			return nil;
	 end

	 local msg = { header }; -- indexed table, starting with header
	 rawMsg = string.sub(rawMsg, header.length+1); -- remove header from raw message

	 -- decompress message if necessary
	 if (header.compression==COMM.HEADER.COMPRESSION.NONE) then
			-- no compression, do nothing
	 elseif (header.compression==COMM.HEADER.COMPRESSION.LIBCOMPRESS) then
			print(string.format("Decompress before: %d",strlen(rawMsg))); -- debug
			rawMsg = Compressor:Decompress(rawMsg);
			print(string.format("Decompress after: %d",strlen(rawMsg))); -- debug
	 else -- no valid compression indicator found, cancel
			-- TODO: Output error message
			return nil;
	 end

	 local messages = { strsplit(COMM.SEP.SECTION, rawMsg) };
	 for i=1, #messages do
			if (messages[i]~=nil and messages[i]~="") then
				 local singleMsg = CommDeserializeMsg(messages[i]);
				 if (singleMsg.type~=COMM.MSG_TYPE.UNKNOWN) then
						tinsert(msg, singleMsg);
				 end
			end
	 end

	 return msg;
end

function BuddyName:OnCommReceived(prefix, msg, channel, sender)
	 if (commActive==false) then -- for debugging purposes
			return;
	 end

	 self:Print("Received message via comm"); -- debug, remove me

	 -- test if message is from different addon or yourself
	 if (prefix~=COMM.SYNC_PREFIX or sender==GetUnitName("player",false)) then
			return;
	 end

	 local decodedMsg = CommSplitMsg(msg);

	 if (decodedMsg==nil) then
			self:Print("Received malformed message by "..sender);
			return;
	 end

	 self:Print("Debug: Received message:");
	 self:Print(allToString(decodedMsg));

	 local header = tremove(decodedMsg,1);
	 if (header.protocol < COMM.PROTOCOL) then
			-- silently ignore old clients for now
			-- perhaps will still be processed later
			-- e.g. by whispering to sender with a hint to update
			return;
	 elseif (header.protocol > COMM.PROTOCOL) then
			if (warnedAboutUpdate==false) then
				 self:Print("A synchronization message was received from a newer client, please update");
				 warnedAboutUpdate=true;
			end
			return;
	 elseif (header.protocol == COMM.PROTOCOL) then

			-- some debug information
			local source;
			if (channel=="GUILD") then
				 source = "guild ("..sender..")";
			elseif (channel=="WHISPER") then
				 source = "player "..sender;
			else
				 source = "Unknown";
			end
			self:Print(string.format("Received synchronization message from %s", source));

			for i=1, #decodedMsg do
				 local currentMsg = decodedMsg[i];
				 self:Print("Processing message of type "..currentMsg.type);

				 if (currentMsg.type == COMM.MSG_TYPE.ENTRY_SET) then
						local entry = currentMsg.entry;

						for j=1, #entry do
							 local currentEntry = entry[j];
							 local charname  = currentEntry[1];
							 local buddyname = currentEntry[2];
							 local timestamp = currentEntry[3];

							 BuddyName:UpdateIfNewer(char,buddy,ts);
						end
				 end
			end
	 end
end

function BuddyName:BroadcastAllToGuild()
	 if (commActive==false) then -- for debugging purposes
			return;
	 end

	 if (PlayerIsInGuild()) then
			local header=COMM.PATTERNS.BUILD.HEADER;

			local msg = COMM.PATTERNS.BUILD.ENTRY_SET_STUB;

			for p,v in pairs(self.db.realm.char) do
				 if (self.db.realm.char[p] and UnitIsInMyGuild(p)) then
						self:Print(string.format("Sharing guild member %s",p)); -- debug
						local newEntry=strjoin(COMM.SEP.FIELD, p, self.db.realm.char[p].buddy,self.db.realm.char[p].timestamp);
						msg=msg..COMM.SEP.ENTRY..newEntry;
				 end
			end

			-- always compress for now
			print(string.format("Compress before: %d",strlen(msg)));
			msg=Compressor:Compress(msg);
			print(string.format("Compress after: %d",strlen(msg)));

			BuddyName:SendCommMessage(COMM.SYNC_PREFIX, msg, "GUILD");
	 end
end

--[[ -- --------------------------------------------- -- ]]
--[[ -- Basic functions automatically called by Ace3  -- ]]
--[[ -- --------------------------------------------- -- ]]

--[[ Executed only once on startup
--   Returns: nil
--]]
function BuddyName:OnInitialize()
	 --self:Print("OnInitialize called"); -- debug

	 currentRealm = GetRealmName();

	 -- Register Database
	 self.db = LibStub("AceDB-3.0"):New("BuddyName_DB")

	 -- Register Slash commands
	 for p,v in pairs(slashCommands) do
			if (v == true) then
				 self:RegisterChatCommand(p, "SlashProcessor")
			end
	 end

	 -- Create frame for events to hook up to and register battle net events
	 BN_EventFrame = CreateFrame("Frame", "BuddyNameFrame");
	 BN_EventFrame:SetScript("OnEvent", eventHandler);
	 for p,v in pairs(registerEvents) do
			if (v == true) then
				 BN_EventFrame:RegisterEvent(p);
			end
	 end

	 -- Register for receiving communication for synchronization purposes
	 BuddyName:RegisterComm(COMM.SYNC_PREFIX);
end


--[[ Executed each time the addon is switched on
--   Returns: nil
--]]
function BuddyName:OnEnable()
	 --self:Print("OnEnable called"); -- debug

	 if (BN_Active == true) then
			return -- we're already active
	 else
			BN_Active = true;
	 end

	 -- check if DB already exists, or if it is old and has to be converted
	 -- not done earlier because has to be called after PLAYER_ENTERING_WORLD (because of UpdateDB())
	 if (self.db.realm.version==nil) then -- if no version number, then either fresh db or very old. In each case, reset realm DB.
			self.db.realm.version=BN_DB_Version -- set current version number
			self.db.realm.char = nil; -- set empty char table
			self.db.realm.char = {};
	 elseif (self.db.realm.version < BN_DB_Version) then -- update old database
			BuddyName:UpdateDB();
	 elseif (self.db.realm.version > BN_DB_Version) then -- database is newer than BuddyName
			self:Print(L["UPDATE_BUDDYNAME"]);
			BN_Active = false;
			return;
	 end

	 -- Apply fixes for broken DB entries
	 if (self.db.realm.version==2) then
			BuddyName:FixDB(2);
	 end

	 -- Hook into AddMessage
	 for i = 1, NUM_CHAT_WINDOWS do
			if (i~=2) then -- exclude combat log, better check against type than number
				 self:RawHook(_G["ChatFrame"..i], "AddMessage", "AddMessage", true);
			end
	 end

	 -- Add one-shot timer to check for battle.net friends in a few seconds
	 --BN_Timer["BattleNet"] = self:ScheduleTimer("AddBattleNetFriends", 5);
end


--[[ Executed each time the addon is switched off
--   Returns: nil
--]]
function BuddyName:OnDisable()
	 --self:Print("OnDisable called");

	 if (BN_Active == false) then
			return -- we're already inactive
	 else
			BN_Active = false;
	 end

	 -- Unhook from AddMessage
	 self:Unhook("AddMessage");

	 -- Cancel battle.net timer if it has not already fired
	 --self:CancelTimer(BN_Timer["BattleNet"], true); -- true: don't raise exception if timer is invalid
end
